@@ -1,27 +1,47 @@
 /*
- * Copyright (c) 1994 David I. Bell
- * Permission is granted to use, distribute, or modify this source,
- * provided that this copyright notice remains intact.
+ * zio - scanf and printf routines for arbitrary precision integers
  *
- * Scanf and printf routines for arbitrary precision integers.
+ * Copyright (C) 1999  David I. Bell
+ *
+ * Calc is open software; you can redistribute it and/or modify it under
+ * the terms of the version 2.1 of the GNU Lesser General Public License
+ * as published by the Free Software Foundation.
+ *
+ * Calc is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+ * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
+ * Public License for more details.
+ *
+ * A copy of version 2.1 of the GNU Lesser General Public License is
+ * distributed with calc under the filename COPYING-LGPL.  You should have
+ * received a copy with calc; if not, write to Free Software Foundation, Inc.
+ * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ *
+ * @(#) $Revision: 29.1 $
+ * @(#) $Id: zio.c,v 29.1 1999/12/14 09:16:17 chongo Exp $
+ * @(#) $Source: /usr/local/src/cmd/calc/RCS/zio.c,v $
+ *
+ * Under source code control:	1993/07/30 19:42:48
+ * File existed as early as:	1993
+ *
+ * Share and enjoy!  :-)	http://reality.sgi.com/chongo/tech/comp/calc/
  */
 
-#include "stdarg.h"
+
+#include <stdio.h>
+#include "config.h"
 #include "zmath.h"
+#include "args.h"
 
 
-#define	OUTBUFSIZE	200		/* realloc size for output buffers */
+#define OUTBUFSIZE	200		/* realloc size for output buffers */
 
-#define	PUTCHAR(ch)		math_chr(ch)
-#define	PUTSTR(str)		math_str(str)
-#define	PRINTF1(fmt, a1)		math_fmt(fmt, a1)
-#define	PRINTF2(fmt, a1, a2)		math_fmt(fmt, a1, a2)
-#define	PRINTF3(fmt, a1, a2, a3)	math_fmt(fmt, a1, a2, a3)
-#define	PRINTF4(fmt, a1, a2, a3, a4)	math_fmt(fmt, a1, a2, a3, a4)
-
-
-long	_outdigits_ = 20;		/* default digits for output */
-int	_outmode_ = MODE_INITIAL;	/* default output mode */
+#define PUTCHAR(ch)		math_chr(ch)
+#define PUTSTR(str)		math_str(str)
+#define PRINTF1(fmt, a1)		math_fmt(fmt, a1)
+#define PRINTF2(fmt, a1, a2)		math_fmt(fmt, a1, a2)
+#define PRINTF3(fmt, a1, a2, a3)	math_fmt(fmt, a1, a2, a3)
+#define PRINTF4(fmt, a1, a2, a3, a4)	math_fmt(fmt, a1, a2, a3, a4)
 
 
 /*
@@ -56,14 +76,14 @@ static long	outbufused;
  * systems with such bogons.
  */
 void
-zio_init()
+zio_init(void)
 {
-    static int done = 0;	/* 1 => routine already called */
+	static int done = 0;	/* 1 => routine already called */
 
-    if (!done) {
-	outfp = stdout;
-	done = 1;
-    }
+	if (!done) {
+		outfp = stdout;
+		done = 1;
+	}
 }
 
 
@@ -72,8 +92,7 @@ zio_init()
  * handle or into a string.
  */
 void
-math_chr(ch)
-	int ch;
+math_chr(int ch)
 {
 	char	*cp;
 
@@ -83,8 +102,10 @@ math_chr(ch)
 	}
 	if (outbufused >= outbufsize) {
 		cp = (char *)realloc(outbuf, outbufsize + OUTBUFSIZE + 1);
-		if (cp == NULL)
+		if (cp == NULL) {
 			math_error("Cannot realloc output string");
+			/*NOTREACHED*/
+		}
 		outbuf = cp;
 		outbufsize += OUTBUFSIZE;
 	}
@@ -97,21 +118,22 @@ math_chr(ch)
  * to a FILE handle or into a string.
  */
 void
-math_str(str)
-	char	*str;
+math_str(char *str)
 {
 	char	*cp;
-	int	len;
+	long	len;
 
 	if (!outputisstring) {
 		fputs(str, outfp);
 		return;
 	}
-	len = strlen(str);
+	len = (long)strlen(str);
 	if ((outbufused + len) > outbufsize) {
 		cp = (char *)realloc(outbuf, outbufsize + len + OUTBUFSIZE + 1);
-		if (cp == NULL)
+		if (cp == NULL) {
 			math_error("Cannot realloc output string");
+			/*NOTREACHED*/
+		}
 		outbuf = cp;
 		outbufsize += (len + OUTBUFSIZE);
 	}
@@ -129,9 +151,7 @@ math_str(str)
  * No characters within the string are handled specially.
  */
 void
-math_fill(str, width)
-	char *str;
-	long width;
+math_fill(char *str, long width)
 {
 	if (width > 0) {
 		width -= strlen(str);
@@ -151,31 +171,13 @@ math_fill(str, width)
  * Routine to output a printf-style formatted string either
  * to a FILE handle or into a string.
  */
-#ifdef VARARGS
-# define VA_ALIST fmt, va_alist
-# define VA_DCL char *fmt; va_dcl
-#else
-# if defined(__STDC__) && __STDC__ != 0
-#  define VA_ALIST char *fmt, ...
-#  define VA_DCL
-# else
-#  define VA_ALIST fmt
-#  define VA_DCL char *fmt;
-# endif
-#endif
-/*VARARGS*/
 void
-math_fmt(VA_ALIST)
-	VA_DCL
+math_fmt(char *fmt, ...)
 {
 	va_list ap;
 	char buf[200];
 
-#ifdef VARARGS
-	va_start(ap);
-#else
 	va_start(ap, fmt);
-#endif
 	vsprintf(buf, fmt, ap);
 	va_end(ap);
 	math_str(buf);
@@ -186,7 +188,7 @@ math_fmt(VA_ALIST)
  * Flush the current output stream.
  */
 void
-math_flush()
+math_flush(void)
 {
 	if (!outputisstring)
 		fflush(outfp);
@@ -196,20 +198,22 @@ math_flush()
 /*
  * Divert further output so that it is saved into a string that will be
  * returned later when the diversion is completed.  The current state of
- * output is remembered for later restoration.  Diversions can be nested.
+ * output is remembered for later restoration.	Diversions can be nested.
  * Output diversion is only intended for saving output to "stdout".
  */
 void
-math_divertio()
+math_divertio(void)
 {
 	register IOSTATE *sp;
 
 	sp = (IOSTATE *) malloc(sizeof(IOSTATE));
-	if (sp == NULL)
+	if (sp == NULL) {
 		math_error("No memory for diverting output");
+		/*NOTREACHED*/
+	}
 	sp->oldiostates = oldiostates;
-	sp->outdigits = _outdigits_;
-	sp->outmode = _outmode_;
+	sp->outdigits = conf->outdigits;
+	sp->outmode = conf->outmode;
 	sp->outfp = outfp;
 	sp->outbuf = outbuf;
 	sp->outbufsize = outbufsize;
@@ -219,8 +223,10 @@ math_divertio()
 	outbufused = 0;
 	outbufsize = 0;
 	outbuf = (char *) malloc(OUTBUFSIZE + 1);
-	if (outbuf == NULL)
+	if (outbuf == NULL) {
 		math_error("Cannot allocate divert string");
+		/*NOTREACHED*/
+	}
 	outbufsize = OUTBUFSIZE;
 	outputisstring = TRUE;
 	oldiostates = sp;
@@ -233,25 +239,28 @@ math_divertio()
  * The string needs freeing by the caller when it is no longer needed.
  */
 char *
-math_getdivertedio()
+math_getdivertedio(void)
 {
 	register IOSTATE *sp;
 	char *cp;
 
 	sp = oldiostates;
-	if (sp == NULL)
+	if (sp == NULL) {
 		math_error("No diverted state to restore");
+		/*NOTREACHED*/
+	}
 	cp = outbuf;
 	cp[outbufused] = '\0';
 	oldiostates = sp->oldiostates;
-	_outdigits_ = sp->outdigits;
-	_outmode_ = sp->outmode;
+	conf->outdigits = sp->outdigits;
+	conf->outmode = sp->outmode;
 	outfp = sp->outfp;
 	outbuf = sp->outbuf;
 	outbufsize = sp->outbufsize;
 	outbufused = sp->outbufused;
 	outbuf = sp->outbuf;
 	outputisstring = sp->outputisstring;
+	free(sp);
 	return cp;
 }
 
@@ -261,7 +270,7 @@ math_getdivertedio()
  * This is called when resetting the global state of the program.
  */
 void
-math_cleardiversions()
+math_cleardiversions(void)
 {
 	while (oldiostates)
 		free(math_getdivertedio());
@@ -279,8 +288,7 @@ math_cleardiversions()
  *	no	no		set output to specified file.
  */
 void
-math_setfp(newfp)
-	FILE *newfp;
+math_setfp(FILE *newfp)
 {
 	outfp = newfp;
 	outputisstring = (oldiostates && (newfp == stdout));
@@ -292,15 +300,16 @@ math_setfp(newfp)
  * This also returns the previous mode.
  */
 int
-math_setmode(newmode)
-	int newmode;
+math_setmode(int newmode)
 {
 	int oldmode;
 
-	if ((newmode <= MODE_DEFAULT) || (newmode > MODE_MAX))
+	if ((newmode <= MODE_DEFAULT) || (newmode > MODE_MAX)) {
 		math_error("Setting illegal output mode");
-	oldmode = _outmode_;
-	_outmode_ = newmode;
+		/*NOTREACHED*/
+	}
+	oldmode = conf->outmode;
+	conf->outmode = newmode;
 	return oldmode;
 }
 
@@ -310,15 +319,16 @@ math_setmode(newmode)
  * This also returns the previous number of digits.
  */
 long
-math_setdigits(newdigits)
-	long newdigits;
+math_setdigits(long newdigits)
 {
 	long olddigits;
 
-	if (newdigits < 0)
+	if (newdigits < 0) {
 		math_error("Setting illegal number of digits");
-	olddigits = _outdigits_;
-	_outdigits_ = newdigits;
+		/*NOTREACHED*/
+	}
+	olddigits = conf->outdigits;
+	conf->outdigits = newdigits;
 	return olddigits;
 }
 
@@ -326,16 +336,14 @@ math_setdigits(newdigits)
 /*
  * Print an integer value as a hex number.
  * Width is the number of columns to print the number in, including the
- * sign if required.  If zero, no extra output is done.  If positive,
+ * sign if required.  If zero, no extra output is done.	 If positive,
  * leading spaces are typed if necessary. If negative, trailing spaces are
- * typed if necessary.  The special characters 0x appear to indicate the
+ * typed if necessary.	The special characters 0x appear to indicate the
  * number is hex.
  */
 /*ARGSUSED*/
 void
-zprintx(z, width)
-	ZVALUE z;
-	long width;
+zprintx(ZVALUE z, long width)
 {
 	register HALF *hp;	/* current word to print */
 	int len;		/* number of halfwords to type */
@@ -353,8 +361,8 @@ zprintx(z, width)
 	if (zisneg(z))
 		PUTCHAR('-');
 	if ((len == 0) && (*z.v <= (HALF) 9)) {
-		len = '0' + *z.v;
-		PUTCHAR(len);
+		len = '0' + (int)(*z.v);
+		PUTCHAR(len & 0xff);
 		return;
 	}
 	hp = z.v + len;
@@ -378,9 +386,7 @@ zprintx(z, width)
  */
 /*ARGSUSED*/
 void
-zprintb(z, width)
-	ZVALUE z;
-	long width;
+zprintb(ZVALUE z, long width)
 {
 	register HALF *hp;	/* current word to print */
 	int len;		/* number of halfwords to type */
@@ -402,20 +408,20 @@ zprintb(z, width)
 	if (zisneg(z))
 		PUTCHAR('-');
 	if ((len == 0) && (*z.v <= (FULL) 1)) {
-		len = '0' + *z.v;
-		PUTCHAR(len);
+		len = '0' + (int)(*z.v);
+		PUTCHAR(len & 0xff);
 		return;
 	}
 	hp = z.v + len;
 	didprint = 0;
 	PUTSTR("0b");
 	while (len-- >= 0) {
-		val = *hp--;
-		mask = (1 << (BASEB - 1));
+		val = ((len >= 0) ? *hp-- : *hp);
+		mask = ((HALF)1 << (BASEB - 1));
 		while (mask) {
 			ch = '0' + ((mask & val) != 0);
 			if (didprint || (ch != '0')) {
-				PUTCHAR(ch);
+				PUTCHAR(ch & 0xff);
 				didprint = 1;
 			}
 			mask >>= 1;
@@ -430,9 +436,7 @@ zprintb(z, width)
  */
 /*ARGSUSED*/
 void
-zprinto(z, width)
-	ZVALUE z;
-	long width;
+zprinto(ZVALUE z, long width)
 {
 	register HALF *hp;	/* current word to print */
 	int len;		/* number of halfwords to type */
@@ -460,8 +464,8 @@ zprinto(z, width)
 		PUTCHAR('-');
 	len = z.len;
 	if ((len == 1) && (*z.v <= (FULL) 7)) {
-		num1 = '0' + *z.v;
-		PUTCHAR(num1);
+		num1 = '0' + (int)(*z.v);
+		PUTCHAR((int)(num1 & 0xff));
 		return;
 	}
 	hp = z.v + len - 1;
@@ -472,10 +476,10 @@ zprinto(z, width)
 		num1 = ((hp[0]) >> 8);
 		num2 = (((hp[0] & 0xff) << 16) + (hp[-1] >> 16));
 		num3 = (((hp[-1] & 0xffff) << 8) + (hp[-2] >> 24));
-		num4 = (hp[-2] % 0xffffff);
+		num4 = (hp[-2] & 0xffffff);
 		if (num1) {
 			PRINTF4("0%lo%08lo%08lo%08lo",
- 			    (PRINT) num1, (PRINT) num2,
+			    (PRINT) num1, (PRINT) num2,
 			    (PRINT) num3, (PRINT) num4);
 		} else {
 			PRINTF3("0%lo%08lo%08lo",
@@ -499,15 +503,17 @@ zprinto(z, width)
 		break;
 	}
 	len -= rem;
-	hp -= rem;
-	while (len > 0) {	/* finish in groups of 3 words */
-		PRINTF4("%08lo%08lo%08lo%08lo",
-		    (PRINT) ((hp[0]) >> 8),
-		    (PRINT) (((hp[0] & 0xff) << 16) + (hp[-1] >> 16)),
-		    (PRINT) (((hp[-1] & 0xffff) << 8) + (hp[-2] >> 24)),
-		    (PRINT) (hp[-2] % 0xffffff));
-		hp -= 3;
-		len -= 3;
+	if (len > 0) {
+		hp -= rem;
+		while (len > 0) {	/* finish in groups of 3 words */
+			PRINTF4("%08lo%08lo%08lo%08lo",
+			    (PRINT) ((hp[0]) >> 8),
+			    (PRINT) (((hp[0] & 0xff) << 16) + (hp[-1] >> 16)),
+			    (PRINT) (((hp[-1] & 0xffff) << 8) + (hp[-2] >> 24)),
+			    (PRINT) (hp[-2] & 0xffffff));
+			hp -= 3;
+			len -= 3;
+		}
 	}
 #else
 	switch (rem) {	/* handle odd amounts first */
@@ -531,13 +537,15 @@ zprinto(z, width)
 		PRINTF1("0%lo", num2);
 	}
 	len -= rem;
-	hp -= rem;
-	while (len > 0) {	/* finish in groups of 3 halfwords */
-		PRINTF2("%08lo%08lo",
-		    ((((FULL) hp[0]) << 8) + (((FULL) hp[-1]) >> 8)),
-		    ((((FULL) (hp[-1] & 0xff)) << 16) + ((FULL) hp[-2])));
-		hp -= 3;
-		len -= 3;
+	if (len > 0) {
+		hp -= rem;
+		while (len > 0) {	/* finish in groups of 3 halfwords */
+			PRINTF2("%08lo%08lo",
+			    ((((FULL) hp[0]) << 8) + (((FULL) hp[-1]) >> 8)),
+			    ((((FULL) (hp[-1] & 0xff))<<16) + ((FULL) hp[-2])));
+			hp -= 3;
+			len -= 3;
+		}
 	}
 #endif
 }
@@ -550,16 +558,18 @@ zprinto(z, width)
  * Decimals supplies number of decimal places to print, with a decimal
  * point at the right location, with zero meaning no decimal point.
  * Width is the number of columns to print the number in, including the
- * decimal point and sign if required.  If zero, no extra output is done.
+ * decimal point and sign if required.	If zero, no extra output is done.
  * If positive, leading spaces are typed if necessary. If negative, trailing
  * spaces are typed if necessary.  As examples of the effects of these values,
- * (345,0,0) = "345", (345,2,0) = "3.45", (345,5,8) = "  .00345".
+ * (345,0,0) = "345", (345,2,0) = "3.45", (345,5,8) = "	 .00345".
+ *
+ * given:
+ *	z		number to be printed
+ *	decimals	number of decimal places
+ *	width		number of columns to print in
  */
 void
-zprintval(z, decimals, width)
-	ZVALUE z;		/* number to be printed */
-	long decimals;		/* number of decimal places */
-	long width;		/* number of columns to print in */
+zprintval(ZVALUE z, long decimals, long width)
 {
 	int depth;		/* maximum depth */
 	int n;			/* current index into array */
@@ -595,6 +605,7 @@ zprintval(z, decimals, width)
 					 &_tenpowers_[depth]);
 			} else {
 				math_error("cannot compute 10^2^(TEN_MAX+1)");
+				/*NOTREACHED*/
 			}
 		}
 	}
@@ -613,21 +624,21 @@ zprintval(z, decimals, width)
 	for (;;) {
 		while (n < depth) {
 			i = depth - n - 1;
-			zdiv(leftnums[n], _tenpowers_[i], &quo, &rem);
+			zdiv(leftnums[n], _tenpowers_[i], &quo, &rem, 0);
 			if (!ziszero(quo))
 				digits += (1L << i);
 			n++;
 			leftnums[n] = quo;
 			rightnums[n] = rem;
 		}
-		i = leftnums[n].v[0];
+		i = (long)(leftnums[n].v[0]);
 		if (output || i || (n == 0)) {
 			if (!output) {
 				output = TRUE;
-				if (decimals > digits)
-					leadspaces -= decimals;
-				else
+				if (decimals < digits)
 					leadspaces -= digits;
+				else
+					leadspaces -= decimals+conf->leadzero;
 				while (--leadspaces >= 0)
 					PUTCHAR(' ');
 				if (neg)
@@ -635,6 +646,8 @@ zprintval(z, decimals, width)
 				if (decimals) {
 					putpoint = (digits - decimals);
 					if (putpoint <= 0) {
+						if (conf->leadzero)
+							PUTCHAR('0');
 						PUTCHAR('.');
 						while (++putpoint <= 0)
 							PUTCHAR('0');
@@ -643,7 +656,7 @@ zprintval(z, decimals, width)
 				}
 			}
 			i += '0';
-			PUTCHAR(i);
+			PUTCHAR((int)(i & 0xff));
 			if (--putpoint == 0)
 				PUTCHAR('.');
 		}
@@ -664,13 +677,11 @@ zprintval(z, decimals, width)
 /*
  * Read an integer value in decimal, hex, octal, or binary.
  * Hex numbers are indicated by a leading "0x", binary with a leading "0b",
- * and octal by a leading "0".  Periods are skipped over, but any other
+ * and octal by a leading "0".	Periods are skipped over, but any other
  * extraneous character stops the scan.
  */
 void
-atoz(s, res)
-	register char *s;
-	ZVALUE *res;
+str2z(char *s, ZVALUE *res)
 {
 	ZVALUE z, ztmp, digit;
 	HALF digval;
@@ -727,4 +738,32 @@ atoz(s, res)
 	*res = z;
 }
 
+
+void
+fitzprint(ZVALUE z, long digits, long show)
+{
+	ZVALUE ztmp1, ztmp2;
+	long i;
+
+	if (digits <= show) {
+		zprintval(z, 0, 0);
+		return;
+	}
+	show /= 2;
+	ztenpow(digits - show, &ztmp1);
+	(void) zquo(z, ztmp1, &ztmp2, 1);
+	zprintval(ztmp2, 0, 0);
+	zfree(ztmp1);
+	zfree(ztmp2);
+	printf("...");
+	ztenpow(show, &ztmp1);
+	(void) zmod(z, ztmp1, &ztmp2, 0);
+	i = zdigits(ztmp2);
+	while (i++ < show)
+		printf("0");
+	zprintval(ztmp2, 0, 0);
+	zfree(ztmp1);
+	zfree(ztmp2);
+}
+
 /* END CODE */
