@@ -1,8 +1,33 @@
 /*
- * Copyright (c) 1994 David I. Bell
- * Permission is granted to use, distribute, or modify this source,
- * provided that this copyright notice remains intact.
+ * zmul - faster than usual multiplying and squaring routines
  *
+ * Copyright (C) 1999  David I. Bell
+ *
+ * Calc is open software; you can redistribute it and/or modify it under
+ * the terms of the version 2.1 of the GNU Lesser General Public License
+ * as published by the Free Software Foundation.
+ *
+ * Calc is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+ * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
+ * Public License for more details.
+ *
+ * A copy of version 2.1 of the GNU Lesser General Public License is
+ * distributed with calc under the filename COPYING-LGPL.  You should have
+ * received a copy with calc; if not, write to Free Software Foundation, Inc.
+ * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ *
+ * @(#) $Revision: 29.1 $
+ * @(#) $Id: zmul.c,v 29.1 1999/12/14 09:16:17 chongo Exp $
+ * @(#) $Source: /usr/local/src/cmd/calc/RCS/zmul.c,v $
+ *
+ * Under source code control:	1991/01/09 20:01:31
+ * File existed as early as:	1991
+ *
+ * Share and enjoy!  :-)	http://reality.sgi.com/chongo/tech/comp/calc/
+ */
+
+/*
  * Faster than usual multiplying and squaring routines.
  * The algorithm used is the reasonably simple one from Knuth, volume 2,
  * section 4.3.3.  These recursive routines are of speed O(N^1.585)
@@ -11,17 +36,15 @@
  * two algorithms are equal in speed at about 100 decimal digits.
  */
 
+
+#include "config.h"
 #include "zmath.h"
 
 
-LEN _mul2_ = MUL_ALG2;		/* size of number to use multiply algorithm 2 */
-LEN _sq2_ = SQ_ALG2;		/* size of number to use square algorithm 2 */
-
-
 static HALF *tempbuf;		/* temporary buffer for multiply and square */
 
-static LEN domul MATH_PROTO((HALF *v1, LEN size1, HALF *v2, LEN size2, HALF *ans));
-static LEN dosquare MATH_PROTO((HALF *vp, LEN size, HALF *ans));
+static LEN domul(HALF *v1, LEN size1, HALF *v2, LEN size2, HALF *ans);
+static LEN dosquare(HALF *vp, LEN size, HALF *ans);
 
 
 /*
@@ -29,11 +52,14 @@ static LEN dosquare MATH_PROTO((HALF *vp, LEN size, HALF *ans));
  *	(A*S+B)*(C*S+D) = (S^2+S)*A*C + S*(A-B)*(D-C) + (S+1)*B*D
  * where S is a power of 2^16, and so multiplies by it are shifts, and
  * A,B,C,D are the left and right halfs of the numbers to be multiplied.
+ *
+ * given:
+ *	z1		numbers to multiply
+ *	z2		numbers to multiply
+ *	res		result of multiplication
  */
 void
-zmul(z1, z2, res)
-	ZVALUE z1, z2;		/* numbers to multiply */
-	ZVALUE *res;		/* result of multiplication */
+zmul(ZVALUE z1, ZVALUE z2, ZVALUE *res)
 {
 	LEN len;		/* size of array */
 
@@ -55,10 +81,10 @@ zmul(z1, z2, res)
 	/*
 	 * Allocate a temporary buffer for the recursion levels to use.
 	 * An array needs to be allocated large enough for all of the
-	 * temporary results to fit in.  This size is about twice the size
+	 * temporary results to fit in.	 This size is about twice the size
 	 * of the largest original number, since each recursion level uses
 	 * the size of its given number, and whose size is 1/2 the size of
-	 * the previous level.  The sum of the infinite series is 2.
+	 * the previous level.	The sum of the infinite series is 2.
 	 * Add some extra words because of rounding when dividing by 2
 	 * and also because of the extra word that each multiply needs.
 	 */
@@ -82,14 +108,16 @@ zmul(z1, z2, res)
  * Returns the actual size of the result with leading zeroes stripped.
  * This also uses a temporary array which must be twice as large as
  * one more than the size of the number at the top level recursive call.
+ *
+ * given:
+ *	v1		first number
+ *	size1		size of first number
+ *	v2		second number
+ *	size2		size of second number
+ *	ans		location for result
  */
 static LEN
-domul(v1, size1, v2, size2, ans)
-	HALF *v1;		/* first number */
-	LEN size1;		/* size of first number */
-	HALF *v2;		/* second number */
-	LEN size2;		/* size of second number */
-	HALF *ans;		/* location for result */
+domul(HALF *v1, LEN size1, HALF *v2, LEN size2, HALF *ans)
 {
 	LEN shift;		/* amount numbers are shifted by */
 	LEN sizeA;		/* size of left half of first number */
@@ -157,10 +185,10 @@ domul(v1, size1, v2, size2, ans)
 	/*
 	 * If the smaller number has only a few digits, then calculate
 	 * the result in the normal manner in order to avoid the overhead
-	 * of the recursion for small numbers.  The number of digits where
+	 * of the recursion for small numbers.	The number of digits where
 	 * the algorithm changes is settable from 2 to maxint.
 	 */
-	if (size2 < _mul2_) {
+	if (size2 < conf->mul2) {
 		/*
 		 * First clear the top part of the result, and then multiply
 		 * by the lowest digit to get the first partial sum.  Later
@@ -180,7 +208,7 @@ domul(v1, size1, v2, size2, ans)
 			len -= 4;
 			sival.ivalue = ((FULL) *h1++) * digit + carry;
 			/* ignore Saber-C warning #112 - get ushort from uint */
-			/* 	  ok to ignore on name domul`sival */
+			/*	  ok to ignore on name domul`sival */
 			*hd++ = sival.silow;
 			carry = sival.sihigh;
 			sival.ivalue = ((FULL) *h1++) * digit + carry;
@@ -260,8 +288,8 @@ domul(v1, size1, v2, size2, ans)
 	/*
 	 * Need to multiply by a large number.
 	 * Allocate temporary space for calculations, and calculate the
-	 * value for the shift.  The shift value is 1/2 the size of the
-	 * larger (first) number (rounded up).  The amount of temporary
+	 * value for the shift.	 The shift value is 1/2 the size of the
+	 * larger (first) number (rounded up).	The amount of temporary
 	 * space needed is twice the size of the shift, plus one more word
 	 * for the multiply to use.
 	 */
@@ -276,44 +304,7 @@ domul(v1, size1, v2, size2, ans)
 	 */
 	baseA = v1 + shift;
 	baseB = v1;
-	/*
-	 * XXX - Saber-C Version 3.1 says:
-	 *
-	 *    W#26, Storing a bad pointer into auto variable dmul`baseC.
-	 *
-	 * This warning is issued during the regression test #026
-	 * (read cryrand).
-	 *
-	 * Saver-C claims that v2+shift is past the end of allocated
-	 * memory for v2.
-	 *
-	 * This warning may be triggered by executing the following code:
-	 *
-	 *	a = 0xffff0000ffffffff00000000ffff0000000000000000ffff;
-	 *	config("mul2", 2);
-	 *	pmod(3,a-1,a);
-	 *
-	 * When this code is executed, shift == 6 and v2 is 3 shorts
-	 * long (size2 == 2).  This baseC points 3 shorts beyond the
-	 * allocated end of v2.
-	 *
-	 * The stack was as follows:
-	 *
-	 *	domul(v1=0x2d93d8, size1=12,
-	 *	      v2=0x2ded30, size2=2, ans=0x2ee8a8) at "zmul.c":313
-	 *	zmul(z1=0x2ee928, z2=0x2ee92c, res=0x16d8c0) at "zmul.c":73
-	 *	zpowermod(z1=0x2ee828, z2=0x2ee82c,
-	 *		  z3=0x2ee830, res=0x57bfe4) at "zmod.c":666
-	 *	qpowermod(q1=0x57bf90, q2=0x57bfc8, q3=0x57bf3c) at "qfunc.c":78
-	 *	builtinfunc(...) at "func.c":400
-	 *	o_call(...) at "opcodes.c":2094
-	 *	calculate(...) at "opcodes.c":288
-	 *	evaluate(...) at "codegen.c":170
-	 *	getcommands(...) at "codegen.c":109
-	 *	main(...) at "calc.c":167
-	 */
-	/* 	ok to ignore on name domul`baseC */
-	baseC = v2 + shift;
+	baseC = v2 + ((shift <= size2)	? shift : size2);
 	baseD = v2;
 	baseAB = ans;
 	baseDC = ans + shift;
@@ -399,11 +390,11 @@ domul(v1, size1, v2, size2, ans)
 	 * Note: step (*) above can produce a result which is larger than
 	 * the final product will be, and this is where the extra word
 	 * needed in the product comes from.  After the final subtraction is
-	 * done, the result fits in the expected size.  Using the extra word
+	 * done, the result fits in the expected size.	Using the extra word
 	 * is easier than suppressing the carries and borrows everywhere.
 	 *
 	 * Begin by forming the product (A-B)*(D-C) into a temporary
-	 * location that we save until the final step.  Do each subtraction
+	 * location that we save until the final step.	Do each subtraction
 	 * at positions 0 and S.  Be very careful about the relative sizes
 	 * of the numbers since this result can be negative.  For the first
 	 * step calculate the absolute difference of A and B into a temporary
@@ -439,12 +430,12 @@ domul(v1, size1, v2, size2, ans)
 	carry = 0;
 	while (subsize--) {
 		sival.ivalue = BASE1 - ((FULL) *h1++) + ((FULL) *h2++) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 	while (copysize--) {
 		sival.ivalue = (BASE1 - ((FULL) *h1++)) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 
@@ -455,7 +446,7 @@ domul(v1, size1, v2, size2, ans)
 	}
 
 	/*
-	 * This completes the calculation of abs(A-B).  For the next step
+	 * This completes the calculation of abs(A-B).	For the next step
 	 * calculate the absolute difference of D and C into a temporary
 	 * location at position S of the result.  Negate the sign if C is
 	 * larger than D.
@@ -470,8 +461,7 @@ domul(v1, size1, v2, size2, ans)
 			h2--;
 		}
 	}
-	if ((sizeC > sizeD) || ((sizeC == sizeD) && (*h1 > *h2)))
-	{
+	if ((sizeC > sizeD) || ((sizeC == sizeD) && (*h1 > *h2))) {
 		neg = !neg;
 		h1 = baseC;
 		h2 = baseD;
@@ -489,12 +479,12 @@ domul(v1, size1, v2, size2, ans)
 	carry = 0;
 	while (subsize--) {
 		sival.ivalue = BASE1 - ((FULL) *h1++) + ((FULL) *h2++) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 	while (copysize--) {
 		sival.ivalue = (BASE1 - ((FULL) *h1++)) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 	hd = &baseDC[sizeDC - 1];
@@ -504,7 +494,7 @@ domul(v1, size1, v2, size2, ans)
 	}
 
 	/*
-	 * This completes the calculation of abs(D-C).  Now multiply
+	 * This completes the calculation of abs(D-C).	Now multiply
 	 * together abs(A-B) and abs(D-C) into a temporary location,
 	 * which is preserved until the final steps.
 	 */
@@ -615,12 +605,12 @@ domul(v1, size1, v2, size2, ans)
 		while (len--) {
 			sival.ivalue = BASE1 - ((FULL) *hd) +
 				((FULL) *h1++) + carry;
-			*hd++ = BASE1 - sival.silow;
+			*hd++ = (HALF)(BASE1 - sival.silow);
 			carry = sival.sihigh;
 		}
 		while (carry) {
 			sival.ivalue = BASE1 - ((FULL) *hd) + carry;
-			*hd++ = BASE1 - sival.silow;
+			*hd++ = (HALF)(BASE1 - sival.silow);
 			carry = sival.sihigh;
 		}
 	} else {
@@ -657,8 +647,7 @@ domul(v1, size1, v2, size2, ans)
  * and A and B are the left and right halfs of the number to square.
  */
 void
-zsquare(z, res)
-	ZVALUE z, *res;
+zsquare(ZVALUE z, ZVALUE *res)
 {
 	LEN len;
 
@@ -674,7 +663,7 @@ zsquare(z, res)
 	/*
 	 * Allocate a temporary array if necessary for the recursion to use.
 	 * The array needs to be allocated large enough for all of the
-	 * temporary results to fit in.  This size is about 3 times the
+	 * temporary results to fit in.	 This size is about 3 times the
 	 * size of the original number, since each recursion level uses 3/2
 	 * of the size of its given number, and whose size is 1/2 the size
 	 * of the previous level.  The sum of the infinite series is 3.
@@ -686,11 +675,11 @@ zsquare(z, res)
 	res->sign = 0;
 	res->v = alloc((z.len+2) * 2);
 	/*
-	 * XXX - Without the memset below, Purify reports that dosquare()
+	 * Without the memset below, Purify reports that dosquare()
 	 *	 will read uninitialized memory at the dosquare() line below
 	 *	 the comment:
 	 *
-	 *	XXX - uninitialized memory read without zsquare memset
+	 *		uninitialized memory read (see zsquare)
 	 *
 	 * This problem occurs during regression test #622 and may
 	 * be duplicated by executing:
@@ -710,12 +699,14 @@ zsquare(z, res)
  * enough for the result (size * 2).  Returns the size of the result.
  * This uses a temporary array which must be 3 times as large as the
  * size of the number at the top level recursive call.
+ *
+ * given:
+ *	vp		value to be squared
+ *	size		length of value to square
+ *	ans		location for result
  */
 static LEN
-dosquare(vp, size, ans)
-	HALF *vp;		/* value to be squared */
-	LEN size;		/* length of value to square */
-	HALF *ans;		/* location for result */
+dosquare(HALF *vp, LEN size, HALF *ans)
 {
 	LEN shift;		/* amount numbers are shifted by */
 	LEN sizeA;		/* size of left half of number to square */
@@ -763,7 +754,7 @@ dosquare(vp, size, ans)
 	 * result.  The second case is handled explicitly.  The number of
 	 * digits where the algorithm changes is settable from 2 to maxint.
 	 */
-	if (size < _sq2_) {
+	if (size < conf->sq2) {
 		hd = ans;
 		len = sizetotal;
 		while (len--)
@@ -820,7 +811,7 @@ dosquare(vp, size, ans)
 			digit = ((FULL) *hd);
 			sival.ivalue = digit + digit + carry;
 			/* ignore Saber-C warning #112 - get ushort from uint */
-			/* 	  ok to ignore on name dosquare`sival */
+			/*	  ok to ignore on name dosquare`sival */
 			*hd++ = sival.silow;
 			carry = sival.sihigh;
 		}
@@ -960,7 +951,7 @@ dosquare(vp, size, ans)
 		carry = sival.sihigh;
 	}
 	while (carry) {
-		/* XXX - uninitialized memory read without zsquare memset */
+		/* uninitialized memory read (see zsquare) */
 		sival.ivalue = ((FULL) *hd) + carry;
 		*hd++ = sival.silow;
 		carry = sival.sihigh;
@@ -980,8 +971,7 @@ dosquare(vp, size, ans)
 			h2--;
 		}
 	}
-	if ((sizeA > sizeB) || ((sizeA == sizeB) && (*h1 > *h2)))
-	{
+	if ((sizeA > sizeB) || ((sizeA == sizeB) && (*h1 > *h2))) {
 		h1 = baseA;
 		h2 = baseB;
 		sizeAB = sizeA;
@@ -998,12 +988,12 @@ dosquare(vp, size, ans)
 	carry = 0;
 	while (subsize--) {
 		sival.ivalue = BASE1 - ((FULL) *h1++) + ((FULL) *h2++) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 	while (copysize--) {
 		sival.ivalue = (BASE1 - ((FULL) *h1++)) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 
@@ -1024,12 +1014,12 @@ dosquare(vp, size, ans)
 	carry = 0;
 	while (sizeABAB--) {
 		sival.ivalue = BASE1 - ((FULL) *hd) + ((FULL) *h1++) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 	while (carry) {
 		sival.ivalue = BASE1 - ((FULL) *hd) + carry;
-		*hd++ = BASE1 - sival.silow;
+		*hd++ = (HALF)(BASE1 - sival.silow);
 		carry = sival.sihigh;
 	}
 
@@ -1054,10 +1044,12 @@ dosquare(vp, size, ans)
  * be freed by the caller.  There is only one temporary buffer, and so as to
  * avoid possible conflicts this is only used by the lowest level routines
  * such as divide, multiply, and square.
+ *
+ * given:
+ *	len		required number of HALFs in buffer
  */
 HALF *
-zalloctemp(len)
-	LEN len;		/* required number of HALFs in buffer */
+zalloctemp(LEN len)
 {
 	HALF *hp;
 	static LEN buflen;	/* current length of temp buffer */
@@ -1077,12 +1069,13 @@ zalloctemp(len)
 		buflen = 0;
 		free(bufptr);
 	}
-	hp = (HALF *) malloc(len * sizeof(HALF));
-	if (hp == NULL)
+	/* don't call alloc() because _math_abort_ may not be set right */
+	hp = (HALF *) malloc((len+1) * sizeof(HALF));
+	if (hp == NULL) {
 		math_error("No memory for temp buffer");
+		/*NOTREACHED*/
+	}
 	bufptr = hp;
 	buflen = len;
 	return hp;
 }
-
-/* END CODE */
